from typing import List, Tuple

from .base import Collider
from .vector import Vector2
from .rect import Rect
from .convex import ConvexCollider


def is_ccw(points: List[List[float]]):
    assert len(points) == 3, "must be a triangle"

    return (
        (points[1][0] - points[0][0]) * (points[2][1] - points[0][1])
        - (points[2][0] - points[0][0]) * (points[1][1] - points[0][1])
    ) > 0


def make_ear(shape: List[List[float]], ear: int) -> Tuple[List[float]]:
    l = len(shape)
    return shape[(ear - 1) % l], shape[ear], shape[(ear + 1) % l]


def compf(a: float, b: float, epsilon: float = 0.00001) -> bool:
    return a + epsilon > b and a - epsilon < b


def in_tri(point: List[float], tri: List[List[float]]) -> bool:
    # Barycentric coordinates.
    denom = (tri[1][1] - tri[2][1]) * (tri[0][0] - tri[2][0]) + (
        tri[2][0] - tri[1][0]
    ) * (tri[0][1] - tri[2][1])
    if compf(denom, 0.0):
        return False
    denom = 1.0 / denom

    # alpha = ((p2.y - p3.y)*(p.x - p3.x)  + (p3.x - p2.x)*(p.y - p3.y)) /
    #         ((p2.y - p3.y)*(p1.x - p3.x) + (p3.x - p2.x)*(p1.y - p3.y))
    alpha = denom * (
        (tri[1][1] - tri[2][1]) * (point[0] - tri[2][0])
        + (tri[2][0] - tri[1][0]) * (point[1] - tri[2][1])
    )
    if alpha < 0:
        return False

    # beta = ((p3.y - p1.y)*(p.x - p3.x) + (p1.x - p3.x)*(p.y - p3.y)) /
    #        ((p2.y - p3.y)*(p1.x - p3.x) + (p3.x - p2.x)*(p1.y - p3.y));
    beta = denom * (
        (tri[2][1] - tri[0][1]) * (point[0] - tri[2][0])
        + (tri[0][0] - tri[2][0]) * (point[1] - tri[2][1])
    )
    if beta < 0:
        return False

    return (1.0 - alpha - beta) >= 0


class ConcaveCollider(Collider):
    """
    Creates new concave collider from given list of tuples.

    Example of usage:

    .. code-block:: python

        collider_points = [(3, 3), (5, 3), (5, 4), (4, 4), (4, 5), (5, 5), (5, 6), (3, 6)]
        collider = ConcaveCollider(collider_points)

    """

    def __init__(self, points: List[Tuple[float]]):
        super().__init__()
        self.points = [[x, y] for x, y in points]  # Must be mutable
        self._vertices: List[Vector2] = []
        self._colliders: List[ConvexCollider] = []

        self._setup()

    def _setup(self):
        # Triangulate before setting up vertices
        triangles = []

        shape = self.points[:]
        # Use orientation of the top-left-most vertex.
        left = shape[0]
        index = 0
        for x in range(len(shape)):
            v = shape[x]
            if v[0] < left[0] or (v[0] == left[0] and v[1] < left[1]):
                left = v
                index = x

        o = is_ccw(make_ear(shape, index))

        while len(shape) >= 3:
            reflex = []
            eartip = -1

            # For each vertex in the shape
            for i in range(len(shape)):
                if eartip >= 0:
                    break

                # Create a triangle from vertex to adjacent vertices.
                tri = make_ear(shape, i)

                prev = tri[0]
                curr = tri[1]
                next = tri[2]

                # If polygon's orientation doesn't match that of the triangle,
                # it's definitely a reflex and not an ear.
                if is_ccw(tri) != o:
                    reflex.append(i)
                    continue

                # Test reflex vertices first.
                for x in reflex:
                    # If we are testing ourselves, skip.
                    if shape[x] in tri:
                        continue

                    # If any v in triangle, not ear.
                    elif in_tri(shape[x], tri):
                        break

                else:
                    # No reflexes, so we test all past current vertex.
                    for x in range(i + 2, len(shape)):
                        if shape[x] in tri:
                            continue
                        elif in_tri(shape[x], tri):
                            break

                    # No vertices in the triangle, we are an ear.
                    else:
                        eartip = i

            if eartip == -1:
                break
            triangles.append(make_ear(shape, eartip))
            del shape[eartip]

        # Convert triangles to ConvexColliders
        self._colliders = [ConvexCollider(points) for points in triangles]

        self._setup_vertices()
        self._setup_rect()

    def _setup_vertices(self):
        self._vertices = [Vector2(*p) for p in self.points]

    def _setup_rect(self):
        min_x = self._vertices[0].x
        min_y = self._vertices[0].y
        max_x = self._vertices[0].x
        max_y = self._vertices[0].y
        for vec in self._vertices[1:]:
            min_x = min(vec._x, min_x)
            max_x = max(vec._x, max_x)
            min_y = min(vec._y, min_y)
            max_y = max(vec._y, max_y)

        self._rect = Rect(min_x, min_y, max_x - min_x, max_y - min_y)

    def collide(self, other) -> bool:
        """
        Check collision against the other collider.

        :param other: Other collider to check collision against.
        :type other: Collider
        :return: True if collision happens, False otherwise
        :rtype: bool
        """
        # Concave to concave collision
        if isinstance(other, ConcaveCollider):
            for collider in self._colliders:
                if other.collide(collider):
                    return True
            return False

        # Concave to convex collision
        for collider in self._colliders:
            if collider.collide(other):
                return True
        return False

    def point_collide(self, point: Tuple[float]):
        """
        Check if point is within collider.

        .. warning:: Not implemented. Returns currently always False

        :param point: Point to test
        :type point: Tuple[float/int, float/int]
        :return: True if point is collider, False otherwise
        :rtype: bool
        """
        return False

    def _move(self, dx: float, dy: float):
        for point, vertex in zip(self.points, self._vertices):
            point[0] += dx
            point[1] += dy
            vertex.x += dx
            vertex.y += dy

        # Move sub colliders
        for collider in self._colliders:
            collider._move(dx, dy)

        # Move rect
        super()._move(dx, dy)



from typing import Union, Tuple, List

from .rect import Rect

try:
    import pygame

    HAS_PYGAME = True
except ImportError:
    HAS_PYGAME = False


class Collider:
    def __init__(self):
        self._rect: Union[Rect, None] = None
        self.points: List[List[Union[float, int]]] = []

    def _move(self, dx: float, dy: float):
        self._rect.x += dx
        self._rect.y += dy

    @property
    def x(self) -> float:
        """
        Left coordinate of bounding box.

        :getter: the coordinate
        :setter: the new coordinate
        :type: float, int
        """
        return self._rect.x

    @x.setter
    def x(self, value: float):
        self._move(value - self._rect.x, 0.0)

    @property
    def y(self) -> float:
        """
        Top coordinate of bounding box.

        :getter: the coordinate
        :setter: the new coordinate
        :type: float, int
        """
        return self._rect.y

    @y.setter
    def y(self, value: float):
        self._move(0.0, value - self._rect.y)

    @property
    def w(self) -> float:
        """
        Width of bounding box.

        :getter: the width
        :type: float, int
        """
        return self._rect.w

    @property
    def h(self) -> float:
        """
        Height of bounding box.

        :getter: the height
        :type: float, int
        """
        return self._rect.h

    @property
    def left(self) -> float:
        """
        Left coordinate of bounding box.

        :getter: the coordinate
        :setter: the new coordinate
        :type: float, int
        """
        return self._rect.x

    @left.setter
    def left(self, value: float):
        self._move(value - self._rect.x, 0.0)

    @property
    def top(self) -> float:
        """
        Top coordinate of bounding box.

        :getter: the coordinate
        :setter: the new coordinate
        :type: float, int
        """
        return self._rect.y

    @top.setter
    def top(self, value: float):
        self._move(0.0, value - self._rect.y)

    @property
    def right(self) -> float:
        """
        Right coordinate of bounding box.

        :getter: the coordinate
        :setter: the new coordinate
        :type: float, int
        """
        return self._rect.right

    @right.setter
    def right(self, value: float):
        self._move(value - self._rect.right, 0.0)

    @property
    def bottom(self) -> float:
        """
        Bottom coordinate of bounding box.

        :getter: the coordinate
        :setter: the new coordinate
        :type: float, int
        """
        return self._rect.bottom

    @bottom.setter
    def bottom(self, value: float):
        self._move(0.0, value - self._rect.bottom)

    @property
    def topleft(self) -> Tuple[float]:
        """
        Top left coordinate of bounding box.

        :getter: the coordinate
        :setter: the new coordinates
        :type: tuple(float/int, float/int)
        """
        return self._rect.topleft

    @topleft.setter
    def topleft(self, value: Tuple[float]):
        self._move(value[0] - self._rect.x, value[1] - self._rect.y)

    @property
    def topright(self) -> Tuple[float]:
        """
        Top right coordinate of bounding box.

        :getter: the coordinate
        :setter: the new coordinates
        :type: tuple(float/int, float/int)
        """
        return self._rect.topright

    @topright.setter
    def topright(self, value: Tuple[float]):
        self._move(value[0] - self._rect.right, value[1] - self._rect.y)

    @property
    def bottomleft(self) -> Tuple[float]:
        """
        Bottom left coordinate of bounding box.

        :getter: the coordinate
        :setter: the new coordinates
        :type: tuple(float/int, float/int)
        """
        return self._rect.bottomleft

    @bottomleft.setter
    def bottomleft(self, value: Tuple[float]):
        self._move(value[0] - self._rect.x, value[1] - self._rect.bottom)

    @property
    def bottomright(self) -> Tuple[float]:
        """
        Bottom right coordinate of bounding box.

        :getter: the coordinate
        :setter: the new coordinates
        :type: tuple(float/int, float/int)
        """
        return self._rect.bottomright

    @bottomright.setter
    def bottomright(self, value: Tuple[float]):
        self._move(value[0] - self._rect.right, value[1] - self._rect.bottom)

    @property
    def centerx(self) -> float:
        """
        Center x coordinate of bounding box.

        :getter: the coordinate
        :setter: the new coordinate
        :type: float, int
        """
        return self._rect.centerx

    @centerx.setter
    def centerx(self, value: float):
        self._move(value - self._rect.centerx, 0.0)

    @property
    def centery(self) -> float:
        """
        Center y coordinate of bounding box.

        :getter: the coordinate
        :setter: the new coordinate
        :type: float, int
        """
        return self._rect.centery

    @centery.setter
    def centery(self, value: float):
        self._move(0.0, value - self.centery)

    @property
    def center(self) -> Tuple[float]:
        """
        Center coordinate of bounding box.

        :getter: the coordinates
        :setter: the new coordinates
        :type: tuple(float/int, float/int)
        """
        return self._rect.center

    @center.setter
    def center(self, value: Tuple[float]):
        self._move(value[0] - self._rect.centerx, value[1] - self._rect.centery)

    @property
    def midtop(self) -> Tuple[float]:
        """
        Center top coordinate of bounding box.

        :getter: the coordinate
        :setter: the new coordinates
        :type: tuple(float/int, float/int)
        """
        return self._rect.midtop

    @midtop.setter
    def midtop(self, value: Tuple[float]):
        self._move(value[0] - self._rect.centerx, value[1] - self._rect.y)

    @property
    def midright(self) -> Tuple[float]:
        """
        Right center coordinate of bounding box.

        :getter: the coordinate
        :setter: the new coordinates
        :type: tuple(float/int, float/int)
        """
        return self._rect.midright

    @midright.setter
    def midright(self, value: Tuple[float]):
        self._move(value[0] - self.right, value[1] - self.centery)

    @property
    def midbottom(self) -> Tuple[float]:
        """
        Center bottom coordinate of bounding box.

        :getter: the coordinate
        :setter: the new coordinates
        :type: tuple(float/int, float/int)
        """
        return self._rect.midbottom

    @midbottom.setter
    def midbottom(self, value: Tuple[float]):
        self._move(value[0] - self._rect.centerx, value[1] - self._rect.bottom)

    @property
    def midleft(self) -> Tuple[float]:
        """
        Left center coordinate of bounding box.

        :getter: the coordinate
        :setter: the new coordinates
        :type: tuple(float/int, float/int)
        """
        return self._rect.midleft

    @midleft.setter
    def midleft(self, value: Tuple[float]):
        self._move(value[0] - self._rect.left, value[1] - self._rect.centery)

    @property
    def size(self) -> Tuple[float]:
        """
        Size of bounding box.

        :getter: the size
        :type: tuple(float/int, float/int)
        """
        return self._rect.size

    @property
    def width(self) -> float:
        """
        Width of bounding box.

        :getter: the width
        :type: float, int
        """
        return self._rect.width

    @property
    def height(self) -> float:
        """
        Height of bounding box.

        :getter: the height
        :type: float, int
        """
        return self._rect.height

    if HAS_PYGAME:

        def draw(self, surface, color=(255, 255, 255), width=1):
            pygame.draw.polygon(surface, color, self.points, width)



from typing import Union, List, Tuple


class Rect:
    """
    A rectangle to hold simple bounding rectangle around the colliders.

    :param x: x coordinate or tuple of (x,y) coordinates
    :param y: y coordinate or tuple of (width, height)
    :param w: width
    :param h: height
    :type x: int, float, tuple(int/float, int/float)
    :type y: int, float, tuple(int/float, int/float)
    :type w: float, int or None
    :type h: float, int or None
    """

    def __init__(
        self,
        x: Union[List[float], float],
        y: Union[List[float], float],
        w: Union[None, int, float] = None,
        h: Union[None, int, float] = None,
    ):
        if isinstance(x, tuple):
            self._x = x[0]
            self._y = x[1]
            if isinstance(y, tuple):
                self._w = y[0]
                self._h = y[1]
            else:
                raise TypeError("Parameters must be (x,y), (w,h) or x, y, w, h")
        else:
            self._x = x
            self._y = y
            self._w = w
            self._h = h

    @property
    def x(self) -> float:
        """
        x location (left) of rect.

        :getter: the location
        :setter: the new location
        :type: float, int
        """
        return self._x

    @x.setter
    def x(self, value: float):
        self._x = value

    @property
    def y(self) -> float:
        """
        y location (top) of rect.

        :getter: the location
        :setter: the new location
        :type: float, int
        """
        return self._y

    @y.setter
    def y(self, value: float):
        self._y = value

    @property
    def w(self) -> float:
        """
        Width of the rect.

        :getter: the width
        :setter: the new width
        :type: float, int
        """
        return self._w

    @w.setter
    def w(self, value: float):
        self._w = value

    @property
    def h(self) -> float:
        """
        Height of rect.

        :getter: the location
        :setter: the new location
        :type: float, int
        """
        return self._h

    @h.setter
    def h(self, value: float):
        self._h = value

    @property
    def left(self) -> float:
        """
        Left location of rect.

        :getter: the location
        :setter: the new location
        :type: float, int
        """
        return self._x

    @left.setter
    def left(self, value: float):
        self._x = value

    @property
    def top(self) -> float:
        """
        Top location of rect.

        :getter: the location
        :setter: the new location
        :type: float, int
        """
        return self._y

    @top.setter
    def top(self, value: float):
        self._y = value

    @property
    def right(self) -> float:
        """
        Right location of rect.

        :getter: the location
        :setter: the new location
        :type: float, int
        """
        return self._x + self._w

    @right.setter
    def right(self, value: float):
        self._x = value - self._w

    @property
    def bottom(self) -> float:
        """
        Bottom location of rect.

        :getter: the location
        :setter: the new location
        :type: float, int
        """
        return self._y + self._h

    @bottom.setter
    def bottom(self, value: float):
        self._y = value - self._h

    @property
    def topleft(self) -> Tuple[float]:
        """
        Top left location of rect.

        :getter: the location
        :setter: the new location
        :type: tuple(float/int, float/int)
        """
        return self._x, self._y

    @topleft.setter
    def topleft(self, value: Tuple[float]):
        self._x = value[0]
        self._y = value[1]

    @property
    def topright(self) -> Tuple[float]:
        """
        Top right location of rect.

        :getter: the location
        :setter: the new location
        :type: tuple(float/int, float/int)
        """
        return self._x + self._w, self._y

    @topright.setter
    def topright(self, value: Tuple[float]):
        self._x = value[0] - self._w
        self._y = value[1]

    @property
    def bottomleft(self) -> Tuple[float]:
        """
        Bottom left location of rect.

        :getter: the location
        :setter: the new location
        :type: tuple(float/int, float/int)
        """
        return self._x, self._y + self._h

    @bottomleft.setter
    def bottomleft(self, value: Tuple[float]):
        self._x = value[0]
        self._y = value[1] - self._h

    @property
    def bottomright(self) -> Tuple[float]:
        """
        Bottom right location of rect.

        :getter: the location
        :setter: the new location
        :type: tuple(float/int, float/int)
        """
        return self._x + self._w, self._y + self._h

    @bottomright.setter
    def bottomright(self, value: Tuple[float]):
        self._x = value[0] - self._w
        self._y = value[1] - self._y

    @property
    def centerx(self) -> float:
        """
        Center x location of rect.

        :getter: the location
        :setter: the new location
        :type: float, int
        """
        return self._x + self._w / 2.0

    @centerx.setter
    def centerx(self, value: float):
        self._x = value - self._w / 2.0

    @property
    def centery(self) -> float:
        """
        Center y location of rect.

        :getter: the location
        :setter: the new location
        :type: float, int
        """
        return self._y + self._h / 2.0

    @centery.setter
    def centery(self, value: float):
        self._y = value - self._h / 2.0

    @property
    def center(self) -> Tuple[float]:
        """
        Center location of rect.

        :getter: the location
        :setter: the new location
        :type: tuple(float/int, float/int)
        """
        return self._x + self._w / 2.0, self._y + self._h / 2.0

    @center.setter
    def center(self, value: Tuple[float]):
        self._x = value[0] - self._w / 2.0
        self._y = value[1] - self._h / 2.0

    @property
    def midtop(self) -> Tuple[float]:
        """
        Mid top location of rect.

        :getter: the location
        :setter: the new location
        :type: tuple(float/int, float/int)
        """
        return self._x + self._w / 2.0, self._y

    @midtop.setter
    def midtop(self, value: Tuple[float]):
        self._x = value[0] - self._w / 2.0
        self._y = value[1]

    @property
    def midright(self) -> Tuple[float]:
        """
        Mid right location of rect.

        :getter: the location
        :setter: the new location
        :type: tuple(float/int, float/int)
        """
        return self._x + self._w, self.y + self._h / 2.0

    @midright.setter
    def midright(self, value: Tuple[float]):
        self._x = value[0] - self._w
        self._y = value[1] - self._w / 2.0

    @property
    def midbottom(self) -> Tuple[float]:
        """
        Mid bottom location of rect.

        :getter: the location
        :setter: the new location
        :type: tuple(float/int, float/int)
        """
        return self._x + self._w / 2.0, self._y + self._h

    @midbottom.setter
    def midbottom(self, value: Tuple[float]):
        self._x = value[0] - self._w / 2.0
        self._y = value[1] - self._h

    @property
    def midleft(self) -> Tuple[float]:
        """
        Mid left location of rect.

        :getter: the location
        :setter: the new location
        :type: tuple(float/int, float/int)
        """
        return self._x, self._y + self._h / 2.0

    @midleft.setter
    def midleft(self, value: Tuple[float]):
        self._x = value[0]
        self._y = value[1] - self._h / 2.0

    @property
    def size(self) -> Tuple[float]:
        """
        Size of rect.

        :getter: the size
        :setter: the new size
        :type: tuple(float/int, float/int)
        """
        return self._w, self._h

    @size.setter
    def size(self, value: Tuple[float]):
        self._w = value[0]
        self._h = value[1]

    @property
    def width(self) -> float:
        """
        Width of rect.

        :getter: the width
        :setter: the new width
        :type: float, int
        """
        return self._w

    @width.setter
    def width(self, value: float):
        self._w = value

    @property
    def height(self) -> float:
        """
        Height of rect.

        :getter: the height
        :setter: the new height
        :type: float, int
        """
        return self._h

    @height.setter
    def height(self, value: float):
        self._h = value



from typing import Union, List, Tuple


class Vector2:
    """
    A class to handle 2 dimensional vector operations.

    Example of usage:

    .. code-block:: python

        vec_a = Vector2(12.1, 23.4)
        vec_b = Vector2(2.5, 11.73)
        vec_c = vec_b - vec_a

    :param x: x component of the vector
    :param y: y component of the vector
    :type x: float, int
    :type y: float, int
    """

    def __init__(self, x: Union[float, int], y: Union[float, int]):
        self._x = x
        self._y = y

    def __add__(self, other: "Vector2") -> "Vector2":
        x: Union[float, int] = self._x + other._x
        y: Union[float, int] = self._y + other._y
        return Vector2(x, y)

    def __sub__(self, other: "Vector2") -> "Vector2":
        x: Union[float, int] = self._x - other._x
        y: Union[float, int] = self._y - other._y
        return Vector2(x, y)

    def __mul__(self, other: "Vector2") -> Union[float, int]:
        """Returns a dot product"""
        return self._x * other._x + self._y * other._y

    @property
    def x(self) -> Union[float, int]:
        """
        Return x component of the vector.

        :getter: a component value.
        :setter: a new component value.
        :type: float, int
        """
        return self._x

    @x.setter
    def x(self, value: Union[float, int]):
        self._x = value

    @property
    def y(self) -> Union[float, int]:
        """
        Return y component of the vector.

        :getter: a component value.
        :setter: a new component value.
        :type: float, int
        """
        return self._y

    @y.setter
    def y(self, value: Union[float, int]):
        self._y = value

    @property
    def xy(self) -> Tuple[Union[float, int]]:
        """
        Return x and y component of the vector as a tuple.

        :getter: a component value as a tuple.
        :setter: a new component value.
        :type: tuple(float/int, float/int)
        """
        return self._x, self._y

    @xy.setter
    def xy(self, value: Union[List[Union[float, int]], Tuple[Union[float, int]]]):
        self._x, self._y = value



import math
from typing import List, Tuple, Union

from .base import Collider
from .vector import Vector2
from .rect import Rect


class ConvexCollider(Collider):
    """
    Creates new convex collider from given list of tuples.

    Example of usage:

    .. code-block:: python

        collider_points = [(11, 10), (11, 3), (4, 3), (4, 10)]
        collider = ConvexCollider(collider_points)

    """

    def __init__(self, points: List[Tuple[Union[float, int]]]):
        super().__init__()
        self.points = [[x, y] for x, y in points]  # Must be mutable
        self._vertices: List[Vector2] = []
        self._edges: List[Vector2] = []
        self._normals: List[Vector2] = []

        self._setup()

    def _setup(self):
        self._setup_vertices()
        self._setup_edges()
        self._setup_normals()
        self._setup_rect()

    def _setup_vertices(self):
        self._vertices = [Vector2(*p) for p in self.points]

    def _setup_edges(self):
        self._edges = []
        vec_a = self._vertices[0]
        for vec_b in self._vertices[1:]:
            self._edges.append(vec_b - vec_a)
            vec_a = vec_b

    def _setup_normals(self):
        self._normals = [Vector2(-e.y, e.x) for e in self._edges]

    def _setup_rect(self):
        min_x = self._vertices[0].x
        min_y = self._vertices[0].y
        max_x = self._vertices[0].x
        max_y = self._vertices[0].y
        for vec in self._vertices[1:]:
            min_x = min(vec._x, min_x)
            max_x = max(vec._x, max_x)
            min_y = min(vec._y, min_y)
            max_y = max(vec._y, max_y)

        self._rect = Rect(min_x, min_y, max_x - min_x, max_y - min_y)

    @property
    def is_clockwise(self) -> bool:
        """
        Checks if convex collider is clockwise.

        :return: True if clockwise, False otherwise
        :rtype: bool
        """
        sum_: float = 0
        prev = first = self._vertices[0]
        for cur in self._vertices[1:]:
            sum_ += (cur.x - prev.x) * (cur.y + prev.y)
            prev = cur
        # This is required to close the loop.
        sum_ += (first.x - prev.x) * (first.y + prev.y)
        return sum_ > 0

    def collide(self, other) -> bool:
        """
        Check collision against the other collider.

        :param other: Other collider to check collision against.
        :type other: Collider
        :return: True if collision happens, False otherwise
        :rtype: bool
        """
        # a = self, b = other
        from .concave import ConcaveCollider  # Late import due cyclic

        if isinstance(other, ConcaveCollider):
            # Check collision other way around
            return other.collide(self)

        normal_stack = self._vertices[:] + other._vertices[:]
        for normal in normal_stack:
            a_min = math.inf
            a_max = -math.inf
            b_min = math.inf
            b_max = -math.inf
            for v in self._vertices:
                dot = normal * v
                a_min = min(a_min, dot)
                a_max = max(a_max, dot)
            for v in other._vertices:
                dot = normal * v
                b_min = min(b_min, dot)
                b_max = max(b_max, dot)
            # Chained comparison due Cython
            if (b_min < a_min < b_max) or (a_min < b_min < a_max):
                continue
            else:
                return False
        return True

    def _make_bbox(self):
        return [
            (self._rect.left, self._rect.top),
            (self._rect.right, self._rect.top),
            (self._rect.right, self._rect.bottom),
            (self._rect.left, self._rect.bottom),
        ]

    def point_collide(self, point: Tuple[float]) -> bool:
        """
        Check if point is within collider.

        :param point: Point to test
        :type point: Tuple[float/int, float/int]
        :return: True if point is collider, False otherwise
        :rtype: bool
        """
        vertx = [v.x for v in self._vertices]
        verty = [v.y for v in self._vertices]

        nvert = len(self._vertices)

        testx = point[0]
        testy = point[1]
        c = 0
        j = nvert - 1
        for i in range(0, nvert):
            if ((verty[i] > testy) != (verty[j] > testy)) and (
                testx
                < (vertx[j] - vertx[i]) * (testy - verty[i]) / (verty[j] - verty[i])
                + vertx[i]
            ):
                c += 1
            j = i
        if c % 2 == 1:
            return True
        return False

    def _move(self, dx: float, dy: float):
        # Move points and vertices
        for point, vertex in zip(self.points, self._vertices):
            point[0] += dx
            point[1] += dy
            vertex.x += dx
            vertex.y += dy

        # Update rect
        super()._move(dx, dy)

    def __str__(self):
        return f"[{','.join([f'[{x}, {y}]' for x, y in self.points])}]"
